import os
from collections.abc import Generator
from typing import Any

import pytest
from ordeq import IOException
from ordeq_args import EnvironmentVariable


@pytest.fixture
def env() -> Generator[None, Any, None]:
    try:
        os.environ["KEY"] = "MyKey"
        yield
    finally:
        del os.environ["KEY"]


def test_it_loads(env):
    assert EnvironmentVariable("KEY").load() == "MyKey"


def test_it_loads_with_default(env):
    assert EnvironmentVariable("KEY", default="DEFAULT").load() == "MyKey"


def test_it_loads_the_default(env):
    assert (
        EnvironmentVariable("NON_EXISTENT_KEY", default="DEFAULT").load()
        == "DEFAULT"
    )


def test_it_raises_an_error_if_key_does_not_exist(env):
    with pytest.raises(IOException):
        EnvironmentVariable("NON_EXISTENT_KEY").load()


def test_it_saves(env):
    EnvironmentVariable("KEY").save("MySavedKey")
    assert os.environ["KEY"] == "MySavedKey"



import os
from dataclasses import KW_ONLY, dataclass

from ordeq.framework.io import IO


@dataclass(frozen=True)
class EnvironmentVariable(IO[str]):
    """IO used to load and save environment variables. Use:
        - as input, to parameterize the node logic
        - as output, to set an environment variable based on node logic

    Gets and sets `os.environ` on load and save. See [1] for more
    information.

    [1] https://docs.python.org/3/library/os.html#os.environ

    Example in a node:

    ```python
    >>> from ordeq import node
    >>> from ordeq_spark import SparkHiveTable
    >>> import pyspark.sql.functions as F
    >>> from pyspark.sql import DataFrame

    >>> @node(
    ...    inputs=[
    ...         SparkHiveTable(table="my.table"),
    ...         EnvironmentVariable("KEY", default="DEFAULT")
    ...    ],
    ...    outputs=SparkHiveTable(table="my.output"),
    ... )
    ... def transform(df: DataFrame, value: str) -> DataFrame:
    ...     return df.where(F.col("col") == value)

    ```

    When you run `transform` through the CLI as follows:

    ```shell
    export KEY=MyValue
    python {your-entrypoint} run --node transform
    ```

    `MyValue` will be used as `value` in `transform`.

    """

    key: str
    _: KW_ONLY
    default: str | None = None

    def load(self) -> str:
        if self.default:
            return os.environ.get(self.key, self.default)
        return os.environ[self.key]

    def save(self, value: str) -> None:
        os.environ[self.key] = value



import argparse
import sys
from unittest.mock import patch

import pytest
from ordeq import node, run
from ordeq_args import CommandLineArg
from ordeq_common import StringBuffer


def test_it_loads():
    with patch.object(sys, "argv", ["sth", "--name", "Barend"]):
        arg = CommandLineArg("--name").load()
        assert arg == "Barend"


def test_it_loads_int():
    with patch.object(sys, "argv", ["sth", "--num", "4"]):
        arg = CommandLineArg("--num", type=int).load()
        assert arg == 4


def test_it_loads_only_known_arg():
    with patch.object(sys, "argv", ["sth", "--name", "Barend", "--num", "4"]):
        arg = CommandLineArg("--num", type=int).load()
        assert arg == 4


def test_it_loads_flag():
    with patch.object(sys, "argv", ["sth", "-n", "4"]):
        arg = CommandLineArg("-n", type=int).load()
        assert arg == 4


def test_it_loads_positional_arg():
    with patch.object(sys, "argv", ["sth", "4"]):
        arg = CommandLineArg("n", type=int).load()
        assert arg == 4


def test_it_works_boolean_arg_false():
    with patch.object(sys, "argv", ["sth", "4"]):
        arg = CommandLineArg("--yes", action="store_true").load()
        assert not arg


def test_it_works_boolean_arg_true():
    with patch.object(sys, "argv", ["sth", "--yes"]):
        arg = CommandLineArg("--yes", action="store_true").load()
        assert arg


@pytest.mark.parametrize(
    ("args", "expected"),
    [
        (["--arg1", "--arg2"], "and"),
        (["--arg1"], "not and"),
        (["--arg2"], "not and"),
        ([], "not and"),
    ],
)
def test_multi_arg(args: list[str], expected: str):
    with patch.object(sys, "argv", ["sth", *args]):
        CLI = [
            CommandLineArg("--arg1", action="store_true"),
            CommandLineArg("--arg2", action="store_true"),
        ]

        def func(v1: bool, v2: bool) -> str:
            return "and" if v1 and v2 else "not and"

        x = StringBuffer()
        n = node(func, inputs=CLI, outputs=x)
        run(n)
        assert x.load() == expected


def test_multi_arg_help(capsys):
    parser = argparse.ArgumentParser("Test CLI")
    with patch.object(sys, "argv", ["sth", "--help"]):
        CLI = [
            CommandLineArg("--arg1", action="store_true", parser=parser),
            CommandLineArg("--arg2", action="store_true", parser=parser),
        ]

        def func(v1: bool, v2: bool) -> str:
            return "and" if v1 and v2 else "not and"

        x = StringBuffer()
        n = node(func, inputs=CLI, outputs=x)
        with pytest.raises(SystemExit):
            run(n)

        captured = capsys.readouterr().out
        lines = captured.splitlines()
        assert lines[-4] == "options:"
        assert lines[-3] == "  -h, --help  show this help message and exit"
        assert lines[-2] == "  --arg1"
        assert lines[-1] == "  --arg2"
        assert parser.prog == "Test CLI"



import argparse
from typing import Any, TypeVar

from ordeq.framework.io import Input

T = TypeVar("T")


class CommandLineArg(Input[T]):
    """Dataset that represents a command line argument as node input. Useful
    for parameterization of node logic based on arguments in the run command.

    Parses the argument from `sys.argv` on load. See
    [argparse](https://docs.python.org/3/library/argparse.html) for more
    information.

    Example:

    ```python
    >>> from ordeq import node
    >>> from ordeq_spark import SparkHiveTable
    >>> import pyspark.sql.functions as F
    >>> from pyspark.sql import DataFrame

    >>> @node(
    ...    inputs=[
    ...         SparkHiveTable(table="my.table"),
    ...         CommandLineArg("--value")
    ...    ],
    ...    outputs=SparkHiveTable(table="my.output"),
    ... )
    ... def transform(df: DataFrame, value: str) -> DataFrame:
    ...     return df.where(F.col("col") == value)

    ```

    When you run `transform` through the CLI as follows:

    ```shell
    python {your-entrypoint} run --node transform --value MyValue
    ```

    `MyValue` will be used as `value` in `transform`.

    By default, the command line arguments are parsed as string. You can
    parse as different type using built-in type converters, for instance:

    ```python
    >>> K = CommandLineArg("--k", type=int)
    >>> Threshold = CommandLineArg("--threshold", type=float)
    >>> Address = CommandLineArg("--address", type=ascii)
    >>> import pathlib
    >>> Path = CommandLineArg("--path", type=pathlib.Path)
    >>> import datetime
    >>> DateTime = CommandLineArg("--date", type=datetime.date.fromisoformat)

    ```

    Alternatively, you can parse using a user-defined function, e.g.:

        >>> def hyphenated(string: str) -> str:
        ...     return '-'.join([w[:4] for w in string.casefold().split()])
        >>> parser = argparse.ArgumentParser()
        >>> Title = CommandLineArg("--title", type=hyphenated)

    Parsing command line arguments as `argparse.FileType` is discouraged as
    [it has been deprecated](https://docs.python.org/3.14/whatsnew/3.14.html#deprecated)
    from Python 3.14.

    More info on parsing types [here](https://docs.python.org/3/library/argparse.html#type)
    """

    arg: str
    kwargs: dict[str, Any]
    parser: argparse.ArgumentParser

    def __init__(
        self, arg: str, parser: argparse.ArgumentParser | None = None, **kwargs
    ):
        super().__init__()
        self.parser = parser or argparse.ArgumentParser()
        self.parser.add_argument(arg, **kwargs)
        # Storing the argument and kwargs for object description
        self.arg = arg
        self.kwargs = kwargs

    def load(self) -> T:
        ns, _ = self.parser.parse_known_args()
        _, val = vars(ns).popitem()
        return val

    def __repr__(self):
        return f"CommandLineArg(arg={self.arg}, kwargs={self.kwargs})"
